/***************************************************************************
 *             __________               __   ___.
 *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
 *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
 *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
 *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
 *                     \/            \/     \/    \/            \/
 *
 * board.c: Manages board
 *
 * Copyright (C) 2006 Katie Stafford
 *
 * All files in this archive are subject to the GNU General Public License.
 * See the file COPYING in the source tree root for full license agreement.
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 * KIND, either express or implied.
 *
 ****************************************************************************/

#include "board.h"
#include "soundutil.h"
#include "actor.h"
#include "player.h"
#include "imath.h"

static int findgoal_x, findgoal_y;
uint16 *board;
uint8 boardseen[4096] IBSS_ATTR;
uint8 findgoalbuf[4096] IBSS_ATTR;
sint16 boardactor[4096] IBSS_ATTR;
char musname[MAX_PATH];


void findgoal(int x, int y) {
    uint16 stack1[130], stack2[130];
    uint16 *wstack, *rstack;
    uint8 *p;
    uint16 *b;
    int i, j, k, v;

    if (findgoal_x == x && findgoal_y == y)
	return;
    findgoal_x = x;
    findgoal_y = y;

    RB memset(findgoalbuf, 255, 4096);

    wstack = stack2;
    *wstack++ = (x << 6 | y);
    findgoalbuf[x << 6 | y] = 0;

    for (i = 0; i < 128; i++) {
	if (i & 1) {
	    j = wstack - stack1;
	    rstack = stack1;
	    wstack = stack2;
	} else {
	    j = wstack - stack2;
	    rstack = stack2;
	    wstack = stack1;
	}
	if (!j)
	    break;
	while (j--) {
	    k = *rstack++;
	    p = &findgoalbuf[k];
	    b = &board[(4 << 12) | k];
	    if (*p > i) {
		*wstack++ = k;
		continue;
	    }
	    v = i + 1;

	    if (k > 0 && !(b[-1] & 16384) && p[-1] > v) {
		p[-1] = v;
		*wstack++ = k - 1;
	    }
	    if (k < 4095 && !(b[1] & 16384) && p[1] > v) {
		p[1] = v;
		*wstack++ = k + 1;
	    }
	    if (k > 63 && !(b[-64] & 16384) && p[-64] > v) {
		p[-64] = v;
		*wstack++ = k - 64;
	    }
	    if (k < (4096-64) && !(b[64] & 16384) && p[64] > v) {
		p[64] = v;
		*wstack++ = k + 64;
	    }
	}
    }
}

int trace(int sx, int sy, int dx, int dy, int *hx, int *hy) {
    int xtemp, ytemp;
    int vx, vy;
    int xscan, yscan, xdir, ydir, incx, incy;
    int d, thx, thy;

    vx = dx - sx;
    vy = dy - sy;

    xscan = sx >> 16;
    xdir = vx > 0 ? 1 : -1;
    incx = abs(vx);
    yscan = sy >> 16;
    ydir = vy > 0 ? 1 : -1;
    incy = abs(vy);

    xtemp = sx & 65535;
    if (xdir > 0)
	xtemp = 65536 - xtemp;
    ytemp = sy & 65535;
    if (ydir > 0)
	ytemp = 65536 - ytemp;

    d = dotprod(xtemp, ytemp, incy, -incx);

    while (1) {
	if (d < 0) {
	    xscan += xdir;
	    if (BCELL(xscan, yscan) & 16384) {
		thx = xscan << 16;
		if (xdir < 0)
		    thx += 65536;
		if (hx)
		    *hx = thx;
		if (hy)
		    *hy = sy + mul64(vy, thx - sx) / vx;
		return ((thx > sx) == (thx > dx));
	    }
	    d += incy;
	} else {
	    yscan += ydir;
	    if (BCELL(xscan, yscan) & 16384) {
		thy = yscan << 16;
		if (ydir < 0)
		    thy += 65536;
		if (hy)
		    *hy = thy;
		if (hx)
		    *hx = sx + mul64(vx, thy - sy) / vy;
		return ((thy > sy) == (thy > dy));
	    }
	    d -= incx;
	}
    }
}

void board_changed(void) {
    findgoal_x = -1;
    findgoal_y = -1;
}


int loadboard(char *fname) {
    char* tmp;
    int i, result;
    uint16 *pos = NULL;
    buffer boardbuf;

    clear_actors();

    if (board) {
	free(board);
	board = NULL;
    }

    /*music_change(NULL, 8, 0);*/
        

    if (!fname) 
        return 0;
    result = loadgzfile(fname, &boardbuf);
    if (result < 0) 
	return result;

    board = (uint16 *) boardbuf.data;

    if (boardbuf.len < 8192 * 5) {
	free(board);
	board = NULL;
	return -4;
    }


#ifdef ROCKBOX_LITTLE_ENDIAN
    pos = board;
    for (i = 0; i < 4096 * 5; i++, pos++) {
	*pos = (*pos) >> 8 | (*pos) << 8;
    }
#else
    (void)i;
    (void)pos;
#endif

    boardbuf.pos = 8192 * 5;

    while (1) {
	actor* o = actor_create(&boardbuf);
	if (!o) break;
	if (HASTAG(o, is_player)) {
	    if (cplayer) {
		player_copy(cplayer, o);
		object_deref(cplayer);
	    }
	    cplayer = o;
            object_ref(cplayer);
	}
	actor_add(o);
    }

    board = realloc(board, 8192 * 5 + 4);

    tmp = RB strrchr(fname, '/');
    if (!tmp) tmp = fname;
    else tmp++;

    RB strncpy(musname,tmp,MAX_PATH);

    tmp = RB strrchr(musname,'.');
    if (tmp) {
        *tmp = 0;
	music_change(musname, 8, 8, true);
    } else {
        musname[0] = 0;
    }
    return 0;
}


void save_board_to(buffer* buf) {
    int i, ipos = buf->pos;
    uint16 *pos;
    writedata(buf, (uint8*)board, 8192*5);

#ifdef ROCKBOX_LITTLE_ENDIAN
    pos = (uint16*)&buf->data[ipos];
    for (i = 0; i < 4096 * 5; i++, pos++) {
	*pos = (*pos) >> 8 | (*pos) << 8;
    }
#else
    (void)i;
    (void)pos;
    (void)ipos;
#endif


    writestr(buf, musname);
}

void load_board_from(buffer* buf) {
    int i;
    uint16 *pos;

    if (board) {
	free(board);
	board = NULL;
    }
    board = malloc(8192 * 5);

    RB memcpy(board, &buf->data[buf->pos], 8192*5);

#ifdef ROCKBOX_LITTLE_ENDIAN
    pos = board;
    for (i = 0; i < 4096 * 5; i++, pos++) {
	*pos = (*pos) >> 8 | (*pos) << 8;
    }
#else
    (void)i;
    (void)pos;
#endif

    buf->pos += 8192 * 5;
    readstr(buf, musname, MAX_PATH);
    if (*musname) {
        music_change(musname, 8, 8, true);
        music_wait_load();
    }
}
