/***************************************************************************
 *             __________               __   ___.
 *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
 *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
 *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
 *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
 *                     \/            \/     \/    \/            \/
 *
 * monsters.c: Robot and alien
 *
 * Copyright (C) 2006 Katie Stafford
 *
 * All files in this archive are subject to the GNU General Public License.
 * See the file COPYING in the source tree root for full license agreement.
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 * KIND, either express or implied.
 *
 ****************************************************************************/

#include "actor.h"
#include "miscactors.h"
#include "graphics.h"
#include "board.h"
#include "soundutil.h"
#include "player.h"
#include "imath.h"

int monster_clips(actor * me, actor * o) {
    (void)me;
    return HASTAG(o, clipped_by_monster);
}

int monster_move(actor * o, int speed) {
    int bx, by, dx, dy, nx, ny;
    int poss;
    int dist, min;

    findgoal(cplayer->x >> 16, cplayer->y >> 16);
    bx = o->x >> 16;
    by = o->y >> 16;

    min = FINDBUF(bx - 1, by);
    if (FINDBUF(bx + 1, by) < min) min = FINDBUF(bx + 1, by);
    if (FINDBUF(bx, by - 1) < min) min = FINDBUF(bx, by - 1);
    if (FINDBUF(bx, by + 1) < min) min = FINDBUF(bx, by + 1);

    if (min == 255)
	return 0;

    poss = 0;

    if (FINDBUF(bx + 1, by) == min)
	poss |= 1;
    if (FINDBUF(bx - 1, by) == min)
	poss |= 2;
    if (FINDBUF(bx, by + 1) == min)
	poss |= 4;
    if (FINDBUF(bx, by - 1) == min)
	poss |= 8;

    if (!poss) return 0;


    /* can we see the player? */
    if (trace(o->x, o->y, cplayer->x, cplayer->y, NULL, NULL)) {
	int res;
	dx = o->x;
	dy = o->y;
	o->vx = cplayer->x - o->x;
	o->vy = cplayer->y - o->y;
	dist = (int)lisqrt(mul64(o->vx, o->vx) + mul64(o->vy, o->vy));

	if (dist > 8192) {
	    o->vx = mul64(o->vx, speed) / dist;
	    o->vy = mul64(o->vy, speed) / dist;
	} else {
	    o->vx = o->vy = 0;
	    return 0;
	}
	nx = (cplayer->x & ~0xffff) | 32768;
	ny = (cplayer->y & ~0xffff) | 32768;
	res = actor_move(o);
	if (res == 1) {
	    o->x = dx;
	    o->y = dy;
	    o->vx = 0;
	    if (dy < ny) {
		o->vy = speed;
		if (dy + speed > ny)
		    o->vy = ny - dy;
	    } else {
		o->vy = -speed;
		if (dy - speed < ny)
		    o->vy = ny - dy;
	    }
	    res = actor_move(o);
	} else if (res == 2) {
	    o->x = dx;
	    o->y = dy;
	    o->vy = 0;
	    if (dx < nx) {
		o->vx = speed;
		if (dx + speed > nx)
		    o->vx = nx - dx;
	    } else {
		o->vx = -speed;
		if (dx - speed < nx)
		    o->vx = nx - dx;
	    }
	    res = actor_move(o);
	}
	return res;

    }

    nx = o->x;
    ny = o->y;
    dx = (bx << 16) | 32768;
    dy = (by << 16) | 32768;


    if ((poss&3) == 1) {
	dx += 65536;
    } else if ((poss&3) == 2) {
	dx -= 65536;
    }

    if ((poss&12) == 4) {
	dy += 65536;
    } else if ((poss&12) == 8) {
	dy -= 65536;
    }
    
    if (poss == 15) {
	dx += 65536;
    }

    if (nx < dx) {
	nx += speed;
	if (nx > dx)
	    nx = dx;
    }

    if (ny < dy) {
	ny += speed;
	if (ny > dy)
	    ny = dy;
    }

    if (nx > dx) {
	nx -= speed;
	if (nx < dx)
	    nx = dx;
    }

    if (ny > dy) {
	ny -= speed;
	if (ny < dy)
	    ny = dy;
    }
    o->vx = (nx - o->x);
    o->vy = (ny - o->y);

    dist = isqrt(o->vx * o->vx + o->vy * o->vy);

    if (dist > speed) {
	o->vx = div32((o->vx * speed), dist);
	o->vy = div32((o->vy * speed), dist);
    }
    return actor_move(o);
}

/*
 *  Evil-looking robot.
 */

TEXGROUP(robot);
ACTOR_ICLASSDEF(robot, int health, cspeed);

METHOD(int, robot, damage, tagset* type, int amt) {
    if (TAGSET_IN(type, damages_monster)) {
	self->health -= amt;
	if (self->health <= 0) {
	    playsndxy("explo", self->x, self->y, 64);
	    actor_remove(self);
	    addexplosion(self->x, self->y);
	} else {
	    playsndxy("monster_hurt", self->x, self->y, 64);
	}
	return DMG_ABSORB;
    }
    return DMG_NO;
}

METHOD(void, robot, move) {
    int tofs = ((1024-self->cspeed)>>5)-64;

    self->cspeed += randrange(tofs,tofs+64);

    if (self->cspeed < -256) self->cspeed = -256;
    if (self->cspeed > 256) self->cspeed = 256;
    monster_move((actor *) self, (self->cspeed<<3) + 4096);
}

METHOD(void, robot, draw) {
    addsprite(self->x, self->y, 0, robot, 0, DIR_FACE, 0);
}

METHOD(void, robot, touch, actor* o) {
    (void)self;
    CALLMETH(o, damage, TAGSET(damages_player), 1);
}

METHOD(int, robot, clips, actor* o) {
    return monster_clips((actor*)self, o);
}

METHOD(void, robot, load, buffer* b) {
    actor_load(self, b);
    self->w = self->h = 16384;
    self->health = 3;
    self->cspeed = 0;

    ADDTAG(self, monster);
    ADDTAG(self, clipped_by_monster);
}

SERIALIZER(robot) {
    S_INT(health);
    S_INT(cspeed);
}

/*
 *  Alien
 */

static int alien_frame[] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 3, 2, 1 };

TEXGROUP(alien);
ACTOR_ICLASSDEF(alien, int health, cspeed, cframe, hovertime);

METHOD(int, alien, damage, tagset* type, int amt) {
    if (TAGSET_IN(type, damages_monster)) {
	self->health -= amt;
	if (self->health <= 0) {
	    playsndxy("explo", self->x, self->y, 64);
	    actor_remove(self);
	    addexplosion(self->x, self->y);
	} else {
	    playsndxy("monster_hurt", self->x, self->y, 64);
	}
	return DMG_ABSORB;
    }
    return DMG_NO;
}

METHOD(void, alien, move) {
    int tofs = ((1024-self->cspeed)>>5)-64;

    self->cspeed += randrange(tofs,tofs+64);

    self->hovertime += randrange(160,180);

    if (self->cspeed < -256) self->cspeed = -256;
    if (self->cspeed > 256) self->cspeed = 256;
    monster_move((actor *) self, (self->cspeed<<3) + 6500);
}

METHOD(void, alien, draw) {
    self->cframe++;
    if (self->cframe >= 12) self->cframe = 0;
    addsprite(self->x, self->y, isin(self->hovertime)>>4, alien, alien_frame[self->cframe], DIR_FACE, 0);
}

METHOD(void, alien, touch, actor* o) {
    (void)self;
    CALLMETH(o, damage, TAGSET(damages_player), 1);
}

METHOD(int, alien, clips, actor* o) {
    return monster_clips((actor*)self, o);
}

METHOD(void, alien, load, buffer* b) {
    actor_load(self, b);
    self->cframe = randrange(0,11);
    self->w = self->h = 24576;
    self->health = 2;
    self->cspeed = 0;
    self->hovertime = randrange(0,2048);
    ADDTAG(self, monster);
    ADDTAG(self, clipped_by_monster);
}

SERIALIZER(alien) {
    S_INT(health);
    S_INT(cspeed);
    S_INT(cframe);
    S_INT(hovertime);
}
