#!/usr/local/bin/python
############################################################################
#             __________               __   ___.
#   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
#   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
#   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
#   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
#                     \/            \/     \/    \/            \/
#
# gentex.py: Generates rocken.tex from PNG files.
#
# Copyright (C) 2006 Katie Stafford
#
# All files in this archive are subject to the GNU General Public License.
# See the file COPYING in the source tree root for full license agreement.
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
# KIND, either express or implied.
#
############################################################################



import Image,array,struct,gzip
from cStringIO import StringIO as sio

try:
    # Use Psyco if available
    import psyco
    psyco.full()
except:
    pass

tex = [
    # texgroup name, [source textures]
    ('board', ['null', 'checker', 'circles', 'dstripe', 'diamond',
               'doorside', 'worms', 'brick', 'tile', 'wave', 'wavert',
               'wavelf', 'jspenguin']), 
    ('intro', [('intro1', 0, 0, 128, 128),
               ('intro1', 128, 0, 128, 128), 
               ('intro2', 0, 0, 128, 128),
               ('intro2', 128, 0, 128, 128),
               'intro3']),
    ('explosion', ['expl%02d'%d for d in xrange(9)]),
    ('generator', ['gen%02d'%d for d in xrange(3,15)]),
    ('bullet', ['bullet']),
    ('robot', ['bot']),
    ('alien', ['alien%d'%d for d in xrange(9)]),
    ('key', ['key']),
    ('medkit', ['medkit', 'medkitopen']),
    ('door', ['door', 'keydoor']),
    ];

def tex_gray(fname):
    if type(fname) == str:
        print 'reading',fname
        img = Image.open(fname+'.png')
    else:
        fname, sx, sy, sw, sh = fname
        print 'reading',fname,'(%dx%d+%d+%d)'%(sw,sh,sx,sy)
        img = Image.open(fname+'.png').crop((sx, sy, sx+sw, sy+sh))
    s = 0
    sh = 1
    if img.size[0] != img.size[1]:
        print 'foo',img.size
        return 
    
    s = img.size[0]

    while s>>sh:
        sh += 1
    sh -= 1
    if s != (1<<sh):
        print 'foo',s
        return

    dofs = 0;
    dat = [0] * (s*((s + 9) / 10))

    shf = 0

    sx = s
    ex = -1
    for x in xrange(s):
        shf = (x % 10) * 3
        i = (x / 10) << sh
        for y in xrange(s):
            l = img.getpixel((x,y))
            try:
                p, a = l
            except:
                p = l
                a = 255
            if a < 128:
                v = 0
            else:
                if sx > x: sx = x;
                if ex < x: ex = x;
                v = (6 - ((p*7)>>8)) + 1
            dat[i] |= v << shf
            i += 1
    sx -= 1;
    ex += 1
    if sx < 0: sx = 0

    dstart = ((sx)/10)<<sh
    dend = ((ex+9)/10)<<sh

    dat = dat[dstart:dend]
    imgdat = struct.pack(">%dI"%(len(dat)), *dat)


    return sh, imgdat, dstart, sx<<(16-sh), ex<<(16-sh)
            
def tc(nm,*a):
    return nm, [tex_gray(i) for i in a]

tex = [(nm, [tex_gray(i) for i in lst]) for nm, lst in tex]

ptr = 4
for nm,p in tex:
    for s,d,dofs,sx,ex in p:
        ptr += len(d)

f = gzip.GzipFile('../rocken.tex','wb');
f.write(struct.pack(">I",ptr))
for nm,p in tex:
    for s,d,dofs,sx,ex in p:
        f.write(d)

ptr = 4
f.write(struct.pack(">H",len(tex)))
for nm,p in tex:
    f.write(struct.pack(">H",len(nm)))
    f.write(nm);
    f.write(struct.pack(">H",len(p)))
    for s,d,dofs,sx,ex in p:
        f.write(struct.pack(">BIIII",s,ptr,dofs,sx,ex))
        ptr += len(d)
f.write(struct.pack(">H",0))
 
print ptr, ptr/1024.0
